import logging
import re
import sys
import functools
import json

from django.utils import six
try:
    unicode = unicode
except NameError:
    unicode = str

logger = logging.getLogger(__name__)
URL_PARAM_RE = re.compile('(?P<k>[^(=|&)]+)=(?P<v>[^&]+)(&|$)')
URL_PARAM_NO_VALUE_RE = re.compile('(?P<k>[^(&|?)]+)(&|$)')


def import_statsd():
    '''
    Import only the statd by wolph not the mozilla statsd
    TODO: Move to mozilla statds which is more widely used
    '''
    try:
        # check to see if the django_statsd we found
        # supports start (stop) timing.
        import django_statsd
        is_wolphs_statsd = hasattr(
            django_statsd, 'start') and hasattr(django_statsd, 'stop')
        if not is_wolphs_statsd:
            django_statsd = None
    except ImportError:
        django_statsd = None

    return django_statsd

django_statsd = import_statsd()


def start_statsd(path):
    '''
    Simple wrapper to save some typing
    '''
    if django_statsd:
        django_statsd.start(path)


def stop_statsd(path):
    if django_statsd:
        django_statsd.stop(path)


def base64_url_decode_php_style(inp):
    '''
    PHP follows a slightly different protocol for base64 url decode.
    For a full explanation see:
    http://stackoverflow.com/questions/3302946/how-to-base64-url-decode-in-python
    and
    http://sunilarora.org/parsing-signedrequest-parameter-in-python-bas
    '''
    import base64
    padding_factor = (4 - len(inp) % 4) % 4
    inp += "=" * padding_factor
    return base64.b64decode(unicode(inp).translate(
        dict(zip(map(ord, u'-_'), u'+/'))))


def encode_params(params_dict):
    '''
    Take the dictionary of params and encode keys and
    values to ascii if it's unicode
    '''
    encoded = [(smart_str(k), smart_str(v)) for k, v in params_dict.items()]
    encoded_dict = dict(encoded)
    return encoded_dict


def smart_str(s, encoding='utf-8', strings_only=False, errors='strict'):
    """
    Adapted from django, needed for urlencoding
    Returns a bytestring version of 's', encoded as specified in 'encoding'.
    If strings_only is True, don't convert (some) non-string-like objects.
    """
    import types
    if strings_only and isinstance(s, (types.NoneType, int)):
        return s
    elif not isinstance(s, six.string_types):
        try:
            return str(s)
        except UnicodeEncodeError:
            if isinstance(s, Exception):
                # An Exception subclass containing non-ASCII data that doesn't
                # know how to print itself properly. We shouldn't raise a
                # further exception.
                return ' '.join([smart_str(arg, encoding, strings_only,
                                           errors) for arg in s])
            return unicode(s).encode(encoding, errors)
    elif isinstance(s, unicode):
        return s.encode(encoding, errors)
    elif s and encoding != 'utf-8':
        return s.decode('utf-8', errors).encode(encoding, errors)
    else:
        return s


def send_warning(message, request=None, e=None, **extra_data):
    '''
    Uses the logging system to send a message to logging and sentry
    '''
    username = None
    if request and request.user.is_authenticated():
        username = request.user.username

    error_message = None
    if e:
        error_message = unicode(e)

    data = {
        'username': username,
        'body': error_message,
    }
    data.update(extra_data)
    logger.warn(message,
                exc_info=sys.exc_info(), extra={
                    'request': request,
                    'data': data
                })


def merge_urls(generated_url, human_url):
    '''
    merge the generated_url with the human_url following this rules:
    params introduced by generated_url are kept
    final params order comes from generated_url
    there's an hack to support things like this http://url?param&param=value

     >>> gen = "http://mysite.com?p1=a&p2=b&p3=c&p4=d"
     >>> hum = "http://mysite.com?p4=D&p3=C&p2=B"
     >>> merge_urls(gen, hum)
     u'http://mysite.com?p1=a&p2=B&p3=C&p4=D'

     >>> gen = "http://mysite.com?id=a&id_s=b&p_id=d"
     >>> hum = "http://mysite.com?id=A&id_s=B&p_id=D"
     >>> merge_urls(gen, hum)
     u'http://mysite.com?id=A&id_s=B&p_id=D'

     >>> gen = "http://mysite.com?p1=a&p2=b&p3=c&p4=d"
     >>> hum = "http://mysite.com"
     >>> merge_urls(gen, hum)
     u'http://mysite.com'

    >>> gen = "http://ad.zanox.com/ppc/?18595160C2000463397T&zpar4=scrapbook&zpar0=e2494344_c4385641&zpar1=not_authenticated&zpar2=unknown_campaign&zpar3=unknown_ref&ULP=http://www.asos.com/ASOS/ASOS-MARS-Loafer-Shoes/Prod/pgeproduct.aspx?iid=1703516&cid=4172&sh=0&pge=2&pgesize=20&sort=-1&clr=Black&affId=2441"
    >>> hum = "http://ad.zanox.com/ppc/?18595160C2000463397T&zpar3=scrapbook&ULP=http://www.asos.com/ASOS/ASOS-MARS-Loafer-Shoes/Prod/pgeproduct.aspx?iid=1703516&cid=4172&sh=0&pge=2&pgesize=20&sort=-1&clr=Black&affId=2441"
    >>> merge_urls(gen, hum)
    u'http://ad.zanox.com/ppc/?18595160C2000463397T&zpar4=scrapbook&zpar0=e2494344_c4385641&zpar1=not_authenticated&zpar2=unknown_campaign&zpar3=scrapbook&ULP=http://www.asos.com/ASOS/ASOS-MARS-Loafer-Shoes/Prod/pgeproduct.aspx?iid=1703516&cid=4172&sh=0&pge=2&pgesize=20&sort=-1&clr=Black&affId=2441'

    >>> gen = "http://mysite.com?invalidparam&p=2"
    >>> hum = "http://mysite.com?p=1"
    >>> merge_urls(gen, hum)
    u'http://mysite.com?invalidparam&p=1'
    '''
    if '?' not in human_url:
        return u'%s' % human_url

    gen_path, gen_args = generated_url.split('?', 1)
    hum_path, hum_args = human_url.split('?', 1)

    get_args = lambda args: [(m.group('k'), m.group('v'))
                             for m in URL_PARAM_RE.finditer(args)]
    get_novalues_args = lambda args: [m.group('k')
                                      for m in URL_PARAM_NO_VALUE_RE.finditer(
                                          args) if "=" not in m.group('k')]

    hum_dict = dict(get_args(hum_args))

    out_args = []

    # prepend crazy param w/o values
    for param in get_novalues_args(gen_args):
        out_args.append(u'%s' % param)

    # replace gen url params
    for k, v in get_args(gen_args):
        out_args.append(u'%s=%s' % (k, hum_dict.get(k, v)))

    return u'%s?%s' % (gen_path, '&'.join(out_args))


class memoized(object):

    '''Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).
    '''

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        try:
            return self.cache[args]
        except KeyError:
            value = self.func(*args)
            self.cache[args] = value
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args)

    def __repr__(self):
        '''Return the function's docstring.'''
        return self.func.__doc__

    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)


def camel_to_underscore(name):
    '''Convert camelcase style naming to underscore style naming

    e.g. SpamEggs -> spam_eggs '''
    import string
    for c in string.ascii_uppercase:
        name = name.replace(c, '_%c' % c)
    return name.strip('_').lower()


def validate_is_instance(instance, classes):
    '''
    Usage
    validate_is_instance(10, int)
    validate_is_instance('a', (str, unicode))
    '''
    if not isinstance(classes, tuple):
        classes = (classes,)
    correct_instance = isinstance(instance, classes)
    if not correct_instance:
        raise ValueError(
            'Expected instance type %s found %s' % (classes, type(instance)))


def is_json(content):
    '''
    Unfortunately facebook returns 500s which mean they are down
    Or 500s with a nice error message because you use open graph wrong

    So we have to figure out which is which :)
    '''
    try:
        json.loads(content)
        is_json = True
    except:
        is_json = False
    return is_json



from django.http import QueryDict
from django_facebook import settings as facebook_settings


def generate_oauth_url(scope=facebook_settings.FACEBOOK_DEFAULT_SCOPE,
                       next=None, extra_data=None):
    query_dict = QueryDict('', True)
    canvas_page = (next if next is not None else
                   facebook_settings.FACEBOOK_CANVAS_PAGE)
    query_dict.update(dict(client_id=facebook_settings.FACEBOOK_APP_ID,
                           redirect_uri=canvas_page,
                           scope=','.join(scope)))
    if extra_data:
        query_dict.update(extra_data)
    auth_url = 'https://www.facebook.com/dialog/oauth?%s' % (
        query_dict.urlencode(), )
    return auth_url



from django.conf.urls import url
from django.conf import settings

# help autodiscovery a bit
from django_facebook import admin
from .views import connect, disconnect, example
from .example_views import *

urlpatterns = [
    url(r'^connect/$', connect, name='facebook_connect'),
    url(r'^disconnect/$',
        disconnect, name='facebook_disconnect'),
    url(r'^example/$', example, name='facebook_example'),
]

dev_patterns = [
    url(
        r'^lazy_decorator_example/$', lazy_decorator_example,
        name='facebook_lazy_decorator_example'),
    url(r'^decorator_example/$', decorator_example,
        name='facebook_decorator_example'),
    url(
        r'^decorator_example_scope/$', decorator_example_scope,
        name='facebook_decorator_example_scope'),
    url(r'^wall_post/$',
        wall_post, name='facebook_wall_post'),
    url(r'^checkins/$',
        checkins, name='facebook_checkins'),
    url(r'^image_upload/$',
        image_upload, name='facebook_image_upload'),
    url(r'^canvas/$', canvas, name='facebook_canvas'),
    url(r'^page_tab/$',
        page_tab, name='facebook_page_tab'),
    url(r'^open_graph_beta/$', open_graph_beta,
        name='facebook_open_graph_beta'),
    url(r'^remove_og_share/$', remove_og_share,
        name='facebook_remove_og_share'),
]

# when developing enable the example views
if settings.DEBUG or getattr(settings, 'TESTING', False):
    # only enable example views while developing
    urlpatterns += dev_patterns


# putting this here instead of models.py reduces issues with import ordering
if getattr(settings, 'AUTH_PROFILE_MODULE', None) == 'django_facebook.FacebookProfile':
    '''
    If we are using the django facebook profile model, create the model
    and connect it to the user create signal
    '''

    from django.db.models.signals import post_save
    from django_facebook.models import FacebookProfile
    from django_facebook.utils import get_user_model

    # Make sure we create a FacebookProfile when creating a User
    def create_facebook_profile(sender, instance, created, **kwargs):
        if created:
            FacebookProfile.objects.create(user=instance)

    post_save.connect(create_facebook_profile, sender=get_user_model())



# -*- coding: utf-8 -*-
'''

Open Facebook allows you to use Facebook's open graph API with simple python code

**Features**

    * Supported and maintained
    * Tested so people can contribute
    * Facebook exceptions are mapped
    * Logging


**Basic examples**::

    facebook = OpenFacebook(access_token)

    # Getting info about me
    facebook.get('me')

    # Learning some more about fashiolista
    facebook.get('fashiolista')

    # Writing your first comment
    facebook.set('fashiolista/comments', message='I love Fashiolista!')


    # Posting to a users wall
    facebook.set('me/feed', message='check out fashiolista',
                 url='http://www.fashiolista.com')

    # Liking a page
    facebook.set('fashiolista/likes')

    # Getting who likes cocacola
    facebook.set('cocacola/likes')

    # Use fql to retrieve your name
    facebook.fql('SELECT name FROM user WHERE uid = me()')

    # Executing fql in batch
    facebook.batch_fql([
        'SELECT uid, name, pic_square FROM user WHERE uid = me()',
        'SELECT uid, rsvp_status FROM event_member WHERE eid=12345678',
    ])

    # Uploading pictures
    photo_urls = [
        'http://e.fashiocdn.com/images/entities/0/7/B/I/9/0.365x365.jpg',
        'http://e.fashiocdn.com/images/entities/0/5/e/e/r/0.365x365.jpg',
    ]
    for photo in photo_urls:
        print facebook.set('me/feed', message='Check out Fashiolista',
                           picture=photo, url='http://www.fashiolista.com')


**Getting an access token**

Once you get your access token, Open Facebook gives you access to the Facebook API
There are 3 ways of getting a facebook access_token and these are currently
implemented by Django Facebook.

1. code is passed as request parameter and traded for an
    access_token using the api

2. code is passed through a signed cookie and traded for an access_token

3. access_token is passed directly (retrieved through javascript, which
    would be bad security, or through one of the mobile flows.)

If you are looking to develop your own flow for a different framework have a look at
Facebook's documentation:
http://developers.facebook.com/docs/authentication/

Also have a look at the :class:`.FacebookRequired` decorator and :func:`get_persistent_graph` function to
understand the required functionality


**Api docs**:


'''

from django.http import QueryDict
from django.utils import six
from django.utils.http import urlencode
from django_facebook import settings as facebook_settings
from open_facebook import exceptions as facebook_exceptions
from open_facebook.utils import json, encode_params, send_warning, memoized, \
    stop_statsd, start_statsd
import logging

from django_facebook.utils import to_int
import ssl
import re

try:
    # python 2 imports
    from urlparse import urlparse
    from urllib2 import build_opener, HTTPError, URLError
except ImportError:
    # python 3 imports
    from urllib.error import HTTPError, URLError
    from urllib.parse import urlparse
    from urllib.request import build_opener

logger = logging.getLogger(__name__)


# base timeout, actual timeout will increase when requests fail
REQUEST_TIMEOUT = 10
# two retries was too little, sometimes facebook is a bit flaky
REQUEST_ATTEMPTS = 3


class FacebookConnection(object):

    '''
    Shared utility class implementing the parsing
    of Facebook API responses
    '''
    api_url = 'https://graph.facebook.com/'
    # this older url is still used for fql requests
    old_api_url = 'https://api.facebook.com/method/'

    @classmethod
    def request(cls, path='', post_data=None, old_api=False, **params):
        '''
        Main function for sending the request to facebook

        **Example**::
            FacebookConnection.request('me')

        :param path:
            The path to request, examples: /me/friends/, /me/likes/

        :param post_data:
            A dictionary of data to post

        :param parms:
            The get params to include
        '''
        api_base_url = cls.old_api_url if old_api else cls.api_url
        if getattr(cls, 'access_token', None):
            params['access_token'] = cls.access_token
        url = '%s%s?%s' % (api_base_url, path, urlencode(params))
        response = cls._request(url, post_data)
        return response

    @classmethod
    def _request(cls, url, post_data=None, timeout=REQUEST_TIMEOUT,
                 attempts=REQUEST_ATTEMPTS):
        # change fb__explicitly_shared to fb:explicitly_shared
        if post_data:
            post_data = dict(
                (k.replace('__', ':'), v) for k, v in post_data.items())

        logger.info('requesting url %s with post data %s', url, post_data)
        post_request = (post_data is not None or 'method=post' in url)

        if post_request and facebook_settings.FACEBOOK_READ_ONLY:
            logger.info('running in readonly mode')
            response = dict(id=123456789, setting_read_only=True)
            return response

        # nicely identify ourselves before sending the request
        opener = build_opener()
        opener.addheaders = [('User-agent', 'Open Facebook Python')]

        # get the statsd path to track response times with
        path = urlparse(url).path
        statsd_path = path.replace('.', '_')

        # give it a few shots, connection is buggy at times
        timeout_mp = 0
        while attempts:
            # gradually increase the timeout upon failure
            timeout_mp += 1
            extended_timeout = timeout * timeout_mp
            response_file = None
            encoded_params = encode_params(post_data) if post_data else None
            post_string = (urlencode(encoded_params)
                           if post_data else None)
            try:
                start_statsd('facebook.%s' % statsd_path)

                try:
                    response_file = opener.open(
                        url, post_string, timeout=extended_timeout)
                    response = response_file.read().decode('utf8')
                except (HTTPError,) as e:
                    response_file = e
                    response = response_file.read().decode('utf8')
                    # Facebook sents error codes for many of their flows
                    # we still want the json to allow for proper handling
                    msg_format = 'FB request, error type %s, code %s'
                    logger.warn(msg_format, type(e), getattr(e, 'code', None))
                    # detect if its a server or application error
                    server_error = cls.is_server_error(e, response)
                    if server_error:
                        # trigger a retry
                        raise URLError(
                            'Facebook is down %s' % response)
                break
            except (HTTPError, URLError, ssl.SSLError) as e:
                # These are often temporary errors, so we will retry before
                # failing
                error_format = 'Facebook encountered a timeout (%ss) or error %s'
                logger.warn(error_format, extended_timeout, str(e))
                attempts -= 1
                if not attempts:
                    # if we have no more attempts actually raise the error
                    error_instance = facebook_exceptions.convert_unreachable_exception(
                        e)
                    error_msg = 'Facebook request failed after several retries, raising error %s'
                    logger.warn(error_msg, error_instance)
                    raise error_instance
            finally:
                if response_file:
                    response_file.close()
                stop_statsd('facebook.%s' % statsd_path)

        # Faceboook response is either
        # Valid json
        # A string which is a querydict (a=b&c=d...etc)
        # A html page stating FB is having trouble (but that shouldnt reach
        # this part of the code)
        try:
            parsed_response = json.loads(response)
            logger.info('facebook send response %s' % parsed_response)
        except Exception as e:
            # using exception because we need to support multiple json libs :S
            parsed_response = QueryDict(response, True)
            logger.info('facebook send response %s' % parsed_response)

        if parsed_response and isinstance(parsed_response, dict):
            # of course we have two different syntaxes
            if parsed_response.get('error'):
                cls.raise_error(parsed_response['error']['type'],
                                parsed_response['error']['message'],
                                parsed_response['error'].get('code'))
            elif parsed_response.get('error_code'):
                cls.raise_error(parsed_response['error_code'],
                                parsed_response['error_msg'])

        return parsed_response

    @classmethod
    def is_server_error(cls, e, response):
        '''
        Checks an HTTPError to see if Facebook is down or we are using the
        API in the wrong way
        Facebook doesn't clearly distinquish between the two, so this is a bit
        of a hack
        '''
        from open_facebook.utils import is_json
        server_error = False
        if hasattr(e, 'code') and e.code == 500:
            server_error = True

        # Facebook status codes are used for application logic
        # http://fbdevwiki.com/wiki/Error_codes#User_Permission_Errors
        # The only way I know to detect an actual server error is to check if
        # it looks like their error page
        # TODO: think of a better solution....
        error_matchers = [
            '<title>Facebook | Error</title>',
            'Sorry, something went wrong.'
        ]
        is_error_page = all(
            [matcher in response for matcher in error_matchers])
        if is_error_page:
            server_error = True

        # if it looks like json, facebook is probably not down
        if is_json(response):
            server_error = False

        return server_error

    @classmethod
    def raise_error(cls, error_type, message, error_code=None):
        '''
        Lookup the best error class for the error and raise it

        **Example**::

            FacebookConnection.raise_error(10, 'OAuthException')

        :param error_type:
            the error type from the facebook api call

        :param message:
            the error message from the facebook api call

        :param error_code:
            optionally the error code which facebook send
        '''
        default_error_class = facebook_exceptions.OpenFacebookException

        # get the error code
        error_code = error_code or cls.get_code_from_message(message)
        # also see http://fbdevwiki.com/wiki/Error_codes#User_Permission_Errors
        logger.info('Trying to match error code %s to error class', error_code)

        # lookup by error code takes precedence
        error_class = cls.match_error_code(error_code)

        # try to get error class by direct lookup
        if not error_class:
            if not isinstance(error_type, int):
                error_class = getattr(facebook_exceptions, error_type, None)
            if error_class and not issubclass(error_class, default_error_class):
                error_class = None

        # hack for missing parameters
        if 'Missing' in message and 'parameter' in message:
            error_class = facebook_exceptions.MissingParameter

        # hack for Unsupported delete request
        if 'Unsupported delete request' in message:
            error_class = facebook_exceptions.UnsupportedDeleteRequest

        # fallback to the default
        if not error_class:
            error_class = default_error_class

        logger.info('Matched error to class %s', error_class)
        error_message = message
        if error_code:
            # this is handy when adding new exceptions for facebook errors
            error_message = u'%s (error code %s)' % (message, error_code)

        raise error_class(error_message)

    @classmethod
    def get_code_from_message(cls, message):
        # map error classes to facebook error codes
        # find the error code
        error_code = None
        error_code_re = re.compile('\(#(\d+)\)')
        matches = error_code_re.match(message)
        matching_groups = matches.groups() if matches else None
        if matching_groups:
            error_code = to_int(matching_groups[0]) or None

        return error_code

    @classmethod
    def get_sorted_exceptions(cls):
        from open_facebook.exceptions import get_exception_classes
        exception_classes = get_exception_classes()
        exception_classes.sort(key=lambda e: e.range())
        return exception_classes

    @classmethod
    def match_error_code(cls, error_code):
        '''
        Return the right exception class for the error code
        '''
        exception_classes = cls.get_sorted_exceptions()
        error_class = None
        for class_ in exception_classes:
            codes_list = class_.codes_list()
            # match the error class
            matching_error_class = None
            for code in codes_list:
                if isinstance(code, tuple):
                    start, stop = code
                    if error_code and start <= error_code <= stop:
                        matching_error_class = class_
                        logger.info('Matched error on code %s', code)
                elif isinstance(code, (int, six.integer_types)):
                    if int(code) == error_code:
                        matching_error_class = class_
                        logger.info('Matched error on code %s', code)
                else:
                    raise(
                        ValueError, 'Dont know how to handle %s of '
                        'type %s' % (code, type(code)))
            # tell about the happy news if we found something
            if matching_error_class:
                error_class = matching_error_class
                break
        return error_class


class FacebookAuthorization(FacebookConnection):

    '''
    Methods for getting us an access token

    There are several flows we must support
    * js authentication flow (signed cookie)
    * facebook app authentication flow (signed cookie)
    * facebook oauth redirect (code param in url)
    These 3 options need to be converted to an access token

    Also handles several testing scenarios
    * get app access token
    * create test user
    * get_or_create_test_user
    '''
    @classmethod
    def convert_code(cls, code,
                     redirect_uri='http://local.mellowmorning.com:8000/facebook/connect/'):
        '''
        Turns a code into an access token

        **Example**::

            FacebookAuthorization.convert_code(code)

        :param code:
            The code to convert

        :param redirect_uri:
            The redirect uri with which the code was requested

        :returns: dict

        '''
        kwargs = cls._client_info()
        kwargs['code'] = code
        kwargs['redirect_uri'] = redirect_uri
        response = cls.request('oauth/access_token', **kwargs)
        return response

    @classmethod
    def extend_access_token(cls, access_token):
        '''
        https://developers.facebook.com/roadmap/offline-access-removal/
        We can extend the token only once per day
        Normal short lived tokens last 1-2 hours
        Long lived tokens (given by extending) last 60 days

        **Example**::

            FacebookAuthorization.extend_access_token(access_token)

        :param access_token:
            The access_token to extend

        :returns: dict
        '''
        kwargs = cls._client_info()
        kwargs['grant_type'] = 'fb_exchange_token'
        kwargs['fb_exchange_token'] = access_token
        response = cls.request('oauth/access_token', **kwargs)
        return response

    @classmethod
    def _client_info(cls):
        kwargs = dict(client_id=facebook_settings.FACEBOOK_APP_ID)
        kwargs['client_secret'] = facebook_settings.FACEBOOK_APP_SECRET
        return kwargs

    @classmethod
    def parse_signed_data(cls, signed_request,
                          secret=facebook_settings.FACEBOOK_APP_SECRET):
        '''
        Thanks to
        http://stackoverflow.com/questions/3302946/how-to-base64-url-decode-in-python
        and
        http://sunilarora.org/parsing-signedrequest-parameter-in-python-bas
        '''
        from open_facebook.utils import base64_url_decode_php_style, smart_str
        l = signed_request.split('.', 2)
        encoded_sig = l[0]
        payload = l[1]
        from open_facebook.utils import json
        sig = base64_url_decode_php_style(encoded_sig)
        import hmac
        import hashlib
        data = json.loads(base64_url_decode_php_style(payload).decode('utf-8'))

        algo = data.get('algorithm').upper()
        if algo != 'HMAC-SHA256':
            error_format = 'Unknown algorithm we only support HMAC-SHA256 user asked for %s'
            error_message = error_format % algo
            send_warning(error_message)
            logger.error('Unknown algorithm')
            return None
        else:
            expected_sig = hmac.new(smart_str(secret), msg=smart_str(payload),
                                    digestmod=hashlib.sha256).digest()

        if not sig == expected_sig:
            error_format = 'Signature %s didnt match the expected signature %s'
            error_message = error_format % (sig, expected_sig)
            send_warning(error_message)
            return None
        else:
            logger.debug('valid signed request received..')
            return data

    @classmethod
    def get_app_access_token(cls):
        '''
        Get the access_token for the app that can be used for
        insights and creating test users
        application_id = retrieved from the developer page
        application_secret = retrieved from the developer page
        returns the application access_token
        '''
        kwargs = {
            'grant_type': 'client_credentials',
            'client_id': facebook_settings.FACEBOOK_APP_ID,
            'client_secret': facebook_settings.FACEBOOK_APP_SECRET,
        }
        response = cls.request('oauth/access_token', **kwargs)
        return response['access_token']

    @memoized
    @classmethod
    def get_cached_app_access_token(cls):
        '''
        Caches the access token in memory, good for speeding up testing
        '''
        app_access_token = cls.get_app_access_token()
        return app_access_token

    @classmethod
    def create_test_user(cls, app_access_token, permissions=None, name=None):
        '''
        Creates a test user with the given permissions and name

        :param app_access_token:
            The application's access token

        :param permissions:
            The list of permissions to request for the test user

        :param name:
            Optionally specify the name
        '''
        if not permissions:
            permissions = ['read_stream', 'publish_stream',
                           'user_photos,offline_access']
        if isinstance(permissions, list):
            permissions = ','.join(permissions)

        default_name = 'Permissions %s' % permissions.replace(
            ',', ' ').replace('_', '')
        name = name or default_name

        kwargs = {
            'access_token': app_access_token,
            'installed': True,
            'name': name,
            'method': 'post',
            'permissions': permissions,
        }
        path = '%s/accounts/test-users' % facebook_settings.FACEBOOK_APP_ID
        # add the test user data to the test user data class
        test_user_data = cls.request(path, **kwargs)
        test_user_data['name'] = name
        test_user = TestUser(test_user_data)

        return test_user

    @classmethod
    def get_or_create_test_user(cls, app_access_token, name=None, permissions=None, force_create=False):
        '''
        There is no supported way of get or creating a test user
        However
        - creating a test user takes around 5s
        - you an only create 500 test users
        So this slows your testing flow quite a bit.

        This method checks your test users
        Queries their names (stores the permissions in the name)

        '''
        if not permissions:
            permissions = ['read_stream', 'publish_stream', 'publish_actions',
                           'user_photos,offline_access']
        if isinstance(permissions, list):
            permissions = ','.join(permissions)

        # hacking the permissions into the name of the test user
        default_name = 'Permissions %s' % permissions.replace(
            ',', ' ').replace('_', '')
        name = name or default_name

        # retrieve all test users
        test_users = cls.get_test_users(app_access_token)
        user_id_dict = dict([(int(u['id']), u) for u in test_users])
        user_ids = map(str, user_id_dict.keys())

        # use fql to figure out their names
        facebook = OpenFacebook(app_access_token)
        users = facebook.fql('SELECT uid, name FROM user WHERE uid in (%s)' %
                             ','.join(user_ids))
        users_dict = dict([(u['name'], u['uid']) for u in users])
        user_id = users_dict.get(name)

        if force_create and user_id:
            # we need the users access_token, the app access token doesn't
            # always work, seems to be a bug in the Facebook api
            test_user_data = user_id_dict[user_id]
            cls.delete_test_user(test_user_data['access_token'], user_id)
            user_id = None

        if user_id:
            # we found our user, extend the data a bit
            test_user_data = user_id_dict[user_id]
            test_user_data['name'] = name
            test_user = TestUser(test_user_data)
        else:
            # create the user
            test_user = cls.create_test_user(
                app_access_token, permissions, name)

        return test_user

    @classmethod
    def get_test_users(cls, app_access_token):
        kwargs = dict(access_token=app_access_token)
        path = '%s/accounts/test-users' % facebook_settings.FACEBOOK_APP_ID
        # retrieve all test users
        response = cls.request(path, **kwargs)
        test_users = response['data']
        return test_users

    @classmethod
    def delete_test_user(cls, app_access_token, test_user_id):
        kwargs = dict(access_token=app_access_token, method='delete')
        path = '%s/' % test_user_id

        # retrieve all test users
        response = cls.request(path, **kwargs)
        return response

    @classmethod
    def delete_test_users(cls, app_access_token):
        # retrieve all test users
        test_users = cls.get_test_users(app_access_token)
        test_user_ids = [u['id'] for u in test_users]
        for test_user_id in test_user_ids:
            cls.delete_test_user(app_access_token, test_user_id)


class OpenFacebook(FacebookConnection):

    '''
    The main api class, initialize using

    **Example**::

        graph = OpenFacebook(access_token)
        print(graph.get('me'))

    '''

    def __init__(self, access_token=None, prefetched_data=None,
                 expires=None, current_user_id=None, version=None):
        '''
            :param access_token:
                The facebook Access token
        '''
        self.access_token = access_token
        # extra data coming from signed cookies
        self.prefetched_data = prefetched_data

        # store to enable detection for offline usage
        self.expires = expires

        # hook to store the current user id if representing the
        # facebook connection to a logged in user :)
        self.current_user_id = current_user_id

        if version is None:
            version = 'v1.0'
        self.version = version

    def __getstate__(self):
        '''
        Turns the object into something easy to serialize
        '''
        state = dict(
            access_token=self.access_token,
            prefetched_data=self.prefetched_data,
            expires=self.expires,
        )
        return state

    def __setstate__(self, state):
        '''
        Restores the object from the state dict
        '''
        self.access_token = state['access_token']
        self.prefetched_data = state['prefetched_data']
        self.expires = state['expires']

    def is_authenticated(self):
        '''
        Ask facebook if we have access to the users data

        :returns:  bool
        '''
        try:
            me = self.me()
        except facebook_exceptions.OpenFacebookException as e:
            if isinstance(e, facebook_exceptions.OAuthException):
                raise
            me = None
        authenticated = bool(me)
        return authenticated

    def get(self, path, version=None, **kwargs):
        '''
        Make a Facebook API call

        **Example**::

            open_facebook.get('me')
            open_facebook.get('me', fields='id,name')

        :param path:
            The path to use for making the API call

        :returns:  dict
        '''
        version = version or self.version
        kwargs['version'] = version
        response = self.request(path, **kwargs)
        return response

    def get_many(self, *ids, **kwargs):
        '''
        Make a batched Facebook API call
        For multiple ids

        **Example**::

            open_facebook.get_many('me', 'starbucks')
            open_facebook.get_many('me', 'starbucks', fields='id,name')

        :param path:
            The path to use for making the API call

        :returns:  dict
        '''
        kwargs['ids'] = ','.join(ids)
        return self.request(**kwargs)

    def set(self, path, params=None, version=None, **post_data):
        '''
        Write data to facebook

        **Example**::

            open_facebook.set('me/feed', message='testing open facebook')

        :param path:
            The path to use for making the API call

        :param params:
            A dictionary of get params

        :param post_data:
            The kwargs for posting to facebook

        :returns:  dict
        '''
        version = version or self.version
        assert self.access_token, 'Write operations require an access token'
        if not params:
            params = {}
        params['method'] = 'post'

        params['version'] = version
        response = self.request(path, post_data=post_data, **params)
        return response

    def delete(self, path, *args, **kwargs):
        '''
        Delete the given bit of data

        **Example**::

            graph.delete(12345)

        :param path:
            the id of the element to remove

        '''

        kwargs['method'] = 'delete'
        self.request(path, *args, **kwargs)

    def fql(self, query, **kwargs):
        '''
        Runs the specified query against the Facebook FQL API.

        **Example**::

            open_facebook.fql('SELECT name FROM user WHERE uid = me()')

        :param query:
            The query to execute

        :param kwargs:
            Extra options to send to facebook

        :returns:  dict
        '''
        kwargs['q'] = query
        path = 'fql'

        response = self.request(path, **kwargs)

        # return only the data for backward compatability
        return response['data']

    def batch_fql(self, queries_dict):
        '''
        queries_dict a dict with the required queries
        returns the query results in:

        **Example**::

            response = facebook.batch_fql({
                name: 'SELECT uid, name, pic_square FROM user WHERE uid = me()',
                rsvp: 'SELECT uid, rsvp_status FROM event_member WHERE eid=12345678',
            })

            # accessing the results
            response['fql_results']['name']
            response['fql_results']['rsvp']

        :param queries_dict:
            A dictiontary of queries to execute

        :returns: dict
        '''
        query = json.dumps(queries_dict)
        query_results = self.fql(query)
        named_results = dict(
            [(r['name'], r['fql_result_set']) for r in query_results])

        return named_results

    def me(self):
        '''
        Cached method of requesting information about me
        '''
        me = getattr(self, '_me', None)
        if me is None:
            #self._me = me = self.get('me')
            self._me = me = self.get('me', fields="id,name,email,verified")

        return me

    def permissions(self):
        '''
        Shortcut for self.get('me/permissions') with some extra parsing
        to turn it into a dictionary of booleans

        :returns: dict
        '''
        permissions_dict = {}
        try:
            permissions = {}
            permissions_response = self.get('me/permissions')

            # determine whether we're dealing with 1.0 or 2.0+
            for permission in permissions_response.get('data', []):
                # graph api 2.0+, returns multiple dicts with keys 'status' and
                # 'permission'
                if any(value in ['granted', 'declined'] for value in permission.values()):
                    for perm in permissions_response['data']:
                        grant = perm.get('status') == 'granted'
                        name = perm.get('permission')
                        # just in case something goes sideways
                        if grant and name:
                            permissions_dict[name] = grant
                # graph api 1.0, returns single dict as {permission: intval}
                elif any(value in [0, 1, '0', '1'] for value in permission.values()):
                    permissions = permissions_response['data'][0]
                    permissions_dict = dict([(k, bool(int(v)))
                                             for k, v in permissions.items()
                                             if v == '1' or v == 1])
                break
        except facebook_exceptions.OAuthException:
            pass

        return permissions_dict

    def has_permissions(self, required_permissions):
        '''
        Validate if all the required_permissions are currently given
        by the user

        **Example**::

            open_facebook.has_permissions(['publish_actions','read_stream'])

        :param required_permissions:
            A list of required permissions

        :returns: bool
        '''
        permissions_dict = self.permissions()
        # see if we have all permissions
        has_permissions = True
        for permission in required_permissions:
            if permission not in permissions_dict:
                has_permissions = False
        return has_permissions

    def my_image_url(self, size='large'):
        '''
        Returns the image url from your profile
        Shortcut for me/picture

        :param size:
            the type of the image to request, see facebook for available formats

        :returns: string
        '''
        query_dict = QueryDict('', True)
        query_dict['type'] = size
        query_dict['access_token'] = self.access_token

        url = '%sme/picture?%s' % (self.api_url, query_dict.urlencode())
        return url

    def request(self, path='', post_data=None, old_api=False, version=None, **params):
        url = self.get_request_url(path=path, old_api=old_api, version=version,
                                   **params)
        logger.info('requesting url %s', url)
        response = self._request(url, post_data)
        return response

    def get_request_url(self, path='', old_api=False, version=None, **params):
        '''
        Gets the url for the request.
        '''
        api_base_url = self.old_api_url if old_api else self.api_url
        version = version or self.version

        if getattr(self, 'access_token', None):
            params['access_token'] = self.access_token

        if api_base_url.endswith('/'):
            api_base_url = api_base_url[:-1]

        if path and path.startswith('/'):
            path = path[1:]

        url = '/'.join([api_base_url, version, path])
        return '%s?%s' % (url, urlencode(params))


class TestUser(object):

    '''
    Simple wrapper around test users
    '''

    def __init__(self, data):
        self.name = data['name']
        self.id = data['id']
        self.access_token = data['access_token']
        self.data = data

    def graph(self):
        graph = OpenFacebook(self.access_token)
        return graph

    def __repr__(self):
        return 'Test user %s' % self.name



from django.utils.decorators import available_attrs
from django.utils.functional import wraps
from django_facebook import settings as fb_settings
from django_facebook.api import get_persistent_graph, require_persistent_graph
from django_facebook.utils import get_oauth_url, parse_scope, response_redirect, \
    has_permissions, simplify_class_decorator
from open_facebook import exceptions as open_facebook_exceptions
import logging


logger = logging.getLogger(__name__)


class FacebookRequired(object):

    """
    Decorator which makes the view require the given Facebook perms,
    redirecting to the log-in page if necessary.

    Note we don't actually query the permissions, we just try in the view
    and upon a permission error redirect to login_url
    Querying the permissions would slow down things
    """

    def __init__(self, fn, scope=None, canvas=False, page_tab=False, extra_params=None):
        self.fn = fn
        scope = fb_settings.FACEBOOK_DEFAULT_SCOPE if scope is None else scope
        self.scope = scope
        self.scope_list = parse_scope(scope)
        self.canvas = canvas
        self.page_tab = page_tab
        self.extra_params = extra_params
        # canvas pages always need to be csrf excempt
        csrf_exempt = canvas or page_tab
        self.csrf_exempt = csrf_exempt

    def authenticate(self, fn, request, *args, **kwargs):
        '''
        Authenticate the user

        There are three options
        a.) We have permissions, proceed with the view
        b.) We tried getting permissions and failed, abort...
        c.) We are about to ask for permissions
        '''
        redirect_uri = self.get_redirect_uri(request)
        oauth_url = get_oauth_url(
            self.scope_list, redirect_uri, extra_params=self.extra_params)

        graph = get_persistent_graph(request, redirect_uri=redirect_uri)

        # See if we have all permissions
        permissions_granted = has_permissions(graph, self.scope_list)

        if permissions_granted:
            response = self.execute_view(
                fn, request, graph=graph, *args, **kwargs)
        elif request.POST.get('attempt', request.GET.get('attempt')) == '1':
            # Doing a redirect could end up causing infinite redirects
            # If Facebook is somehow not giving permissions
            # Time to show an error page
            response = self.authentication_failed(fn, request, *args, **kwargs)
        else:
            response = self.oauth_redirect(oauth_url, redirect_uri)

        return response

    def get_redirect_uri(self, request):
        '''
        return the redirect uri to use for oauth authorization
        this needs to be the same for requesting and accepting the token
        '''
        if self.canvas:
            redirect_uri = fb_settings.FACEBOOK_CANVAS_PAGE
        else:
            redirect_uri = request.build_absolute_uri()

        # set attempt=1 to prevent endless redirect loops
        if 'attempt=1' not in redirect_uri:
            if '?' not in redirect_uri:
                redirect_uri += '?attempt=1'
            else:
                redirect_uri += '&attempt=1'

        return redirect_uri

    def __call__(self):
        '''
        When the decorator is called like this
            @facebook_required
            The call will receive

        Otherwise it will be like
            @facebook_required(scope=[])
            The init will receive the parameters
        '''
        @wraps(self.fn, assigned=available_attrs(self.fn))
        def wrapped_view(request, *args, **kwargs):
            response = self.authenticate(self.fn, request, *args, **kwargs)
            return response

        wrapped_view.csrf_exempt = self.csrf_exempt
        return wrapped_view

    def oauth_redirect(self, oauth_url, redirect_uri, e=None):
        '''
        Redirect to Facebook's oAuth dialog
        '''
        logger.info(
            u'requesting access with redirect uri: %s, error was %s',
            redirect_uri, e)

        # for internal Facebook pages we should use a script to redirect
        script_redirect = False
        if self.canvas or self.page_tab:
            script_redirect = True

        # redirect using HTTP headers or a script
        response = response_redirect(
            oauth_url, script_redirect=script_redirect)
        return response

    def authentication_failed(self, fn, request, *args, **kwargs):
        '''
        Execute the view but don't pass the graph to indicate we couldn't
        get the right permissions
        '''
        msg = '''\
            Somehow Facebook is not giving us the permissions needed
            Lets cancel instead of endless redirects
        '''
        logger.info(msg)
        response = self.execute_view(fn, request, graph=None, *args, **kwargs)
        return response

    def execute_view(self, view_func, *args, **kwargs):
        try:
            result = view_func(*args, **kwargs)
        except TypeError as e:
            # this might be another error type error, raise it
            # the only way I know to check this is the message :(
            if 'graph' not in str(e):
                raise
            graph = kwargs.pop('graph', None)
            result = view_func(*args, **kwargs)
        return result


# decorators should look like functions :)
facebook_required = simplify_class_decorator(FacebookRequired)


class FacebookRequiredLazy(FacebookRequired):

    """
    Decorator which makes the view require the given Facebook perms,
    redirecting to the log-in page if necessary.

    Based on exceptions instead of a permission check
    Faster, but more prone to bugs

    Use this in combination with require_persistent_graph
    """

    def authenticate(self, fn, request, *args, **kwargs):
        redirect_uri = self.get_redirect_uri(request)
        oauth_url = get_oauth_url(
            self.scope_list, redirect_uri, extra_params=self.extra_params)

        graph = None
        try:
            # call get persistent graph and convert the
            # token with correct redirect uri
            graph = require_persistent_graph(
                request, redirect_uri=redirect_uri)
            # Note we're not requiring a persistent graph here
            # You should require a persistent graph in the view when you start
            # using this
            response = self.execute_view(
                fn, request, graph=graph, *args, **kwargs)
        except open_facebook_exceptions.OpenFacebookException as e:
            permission_granted = has_permissions(graph, self.scope_list)
            if permission_granted:
                # an error if we already have permissions
                # shouldn't have been caught
                # raise to prevent bugs with error mapping to cause issues
                raise
            elif request.POST.get('attempt', request.GET.get('attempt')) == '1':
                # Doing a redirect could end up causing infinite redirects
                # If Facebook is somehow not giving permissions
                # Time to show an error page
                response = self.authentication_failed(
                    fn, request, *args, **kwargs)
            else:
                response = self.oauth_redirect(oauth_url, redirect_uri, e)
        return response


# decorators should look like functions :)
facebook_required_lazy = simplify_class_decorator(FacebookRequiredLazy)



from django.core.management.base import BaseCommand
import logging
import sys

DEFAULT_VERBOSITY = 1
VERBOSITY_LOG_MAP = {
    0: logging.WARN,
    1: logging.INFO,
    2: logging.DEBUG,
    3: logging.DEBUG,
}


class CustomBaseCommand(BaseCommand):

    '''
    Management command with a convenient self.log.info shortcut allowing for
    easy logging

    Listens to the verbosity command given to the management command
    '''

    def __init__(self, session=None):
        self._session = session
        self.verbosity = DEFAULT_VERBOSITY
        BaseCommand.__init__(self)

    def handle(self, *args, **kwargs):
        self.verbosity = int(kwargs.get('verbosity', DEFAULT_VERBOSITY))

    def create_logger(self):
        name = self.__class__.__module__.split('.')[-1]

        logger_name = 'management.commands.%s' % name
        logger = logging.getLogger(logger_name)
        logger.extra = {
            'view': logger_name,
            'data': {
                'command': ' '.join(sys.argv),
            }
        }

        handler = logging.StreamHandler()
        logger.addHandler(handler)
        logger.setLevel(VERBOSITY_LOG_MAP[self.verbosity])
        return logger

    @property
    def log(self):
        if not hasattr(self, 'logger'):
            self.logger = self.create_logger()
        return self.logger



from django import forms


class SignupForm(forms.Form):

    username = forms.CharField()
    email = forms.EmailField()
    first_name = forms.CharField()
    password1 = forms.CharField()

    def clean(self):
        data = self.cleaned_data
        data['username'] = 'Test form'
        data['password1'] = 'password'
        return data

    def save(self):
        pass



from django.db.models.query_utils import Q
from django.core.cache import cache
from django.db import models
import operator
import random
from datetime import timedelta
from django_facebook.utils import compatible_datetime as datetime
from django.contrib.contenttypes.models import ContentType
import logging
from open_facebook.exceptions import OAuthException, UnsupportedDeleteRequest
logger = logging.getLogger(__name__)


class FacebookUserManager(models.Manager):

    def find_users(self, queries, base_queryset=None):
        '''
        Queries, a list of search queries
        Base Queryset, the base queryset in which we are searching
        '''
        if base_queryset is None:
            base_queryset = self.all()
        filters = []
        for query in queries:
            match = Q(
                name__istartswith=query) | Q(name__icontains=' %s' % query)
            filters.append(match)

        users = base_queryset.filter(reduce(operator.and_, filters))

        return users

    def random_facebook_friends(self, user, gender=None, limit=3):
        '''
        Returns a random sample of your FB friends

        Limit = Number of friends
        Gender = None, M or F
        '''
        assert gender in (
            None, 'M', 'F'), 'Gender %s wasnt recognized' % gender

        from django_facebook.utils import get_profile_model
        facebook_cache_key = 'facebook_users_%s' % user.id
        non_members = cache.get(facebook_cache_key)
        profile_class = get_profile_model()
        if not non_members:
            facebook_users = list(
                self.filter(user_id=user.id, gender=gender)[:50])
            facebook_ids = [u.facebook_id for u in facebook_users]
            members = list(profile_class.objects.filter(
                facebook_id__in=facebook_ids).select_related('user'))
            member_ids = [p.facebook_id for p in members]
            non_members = [
                u for u in facebook_users if u.facebook_id not in member_ids]

            cache.set(facebook_cache_key, non_members, 60 * 60)

        random_limit = min(len(non_members), 3)
        random_facebook_users = []
        if random_limit:
            random_facebook_users = random.sample(non_members, random_limit)

        return random_facebook_users


class OpenGraphShareManager(models.Manager):

    def failed(self):
        qs = self.filter(completed_at__isnull=True)
        return qs

    def recently_failed(self):
        from django_facebook import settings as facebook_settings
        now = datetime.now()
        recent_delta = timedelta(
            days=facebook_settings.FACEBOOK_OG_SHARE_RETRY_DAYS)
        recent = now - recent_delta
        failed = self.failed()
        recently_failed = failed.filter(created_at__gte=recent)
        return recently_failed

    def shares_for_instance(self, instance, user):
        content_type = ContentType.objects.get_for_model(instance)
        shares = self.filter(
            user=user,
            object_id=instance.id,
            content_type=content_type,
            completed_at__isnull=False,
            removed_at__isnull=True,
        )
        return shares

    def remove_shares_for_instance(self, content_object, user):
        '''
        Removes all shares for this content_object and user combination
        '''
        shares = self.shares_for_instance(content_object, user)
        shares = shares.filter(
            completed_at__isnull=False, removed_at__isnull=True)
        shares = list(shares[:1000])
        logger.info('found %s shares to remove', len(shares))
        for share in shares:
            logger.info('removed share %s', share)
            try:
                share.remove()
            except (OAuthException, UnsupportedDeleteRequest) as e:
                # oauth exceptions happen when tokens are removed
                # unsupported delete requests when the resource is already
                # removed
                logger.info('removing share failed, got error %s', e)



from django.contrib import messages
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_protect
from django_facebook.api import get_persistent_graph
from django_facebook.decorators import facebook_required_lazy, facebook_required
from django_facebook.utils import next_redirect, parse_signed_request


@facebook_required
def decorator_example(request, graph):
    '''
    Redirects the user to Facebook's oAuth dialog if the permissions
    requested in scope are not present

    default is FACEBOOK_DEFAULT_SCOPE
    '''
    if graph:
        # Proceed to load the users friends, or maybe post on their wall
        return HttpResponse('authorized')
    else:
        # this happens when the user denies the authentication request
        # or when there is a Facebook error
        return HttpResponse('user denied or error')


@facebook_required(scope=['publish_actions', 'user_status'])
def decorator_example_scope(request, graph):
    '''
    Same as above, but with a custom scope
    '''
    if graph:
        # Proceed to load the users friends, or maybe post on their wall
        return HttpResponse('authorized')
    else:
        # this happens when the user denies the authentication request
        # or when there is a Facebook error
        return HttpResponse('user denied or error')


@facebook_required_lazy
def lazy_decorator_example(request, graph):
    '''
    The lazy decorator is faster, but somewhat harder to use
    You have no guarantee that you have the required permissions
    The user will get redirected to the oAuth dialog if your facebook calls
    trigger a Facebook Exception
    '''
    if graph:
        # Proceed to load the users friends, or maybe post on their wall
        # any action triggering a facebook error will cause a redirect
        # to the facebook oAuth dialog
        if request.method == 'POST':
            # requesting friends will fail if we don't have permissions
            # the exception will trigger a redirect to the oAuth dialog
            friends = graph.get('me/friends')

        return HttpResponse('authorized')
    else:
        # this happens when the user denies the authentication request
        return HttpResponse('user denied or error')


@facebook_required(canvas=True)
def canvas(request, graph):
    '''
    Example of a canvas page.
    Canvas pages require redirects to work using javascript instead of http headers
    The facebook required and facebook required lazy decorator abstract this away
    '''
    signed_request_string = request.POST.get('signed_request')

    signed_request = {}
    if signed_request_string:
        signed_request = parse_signed_request(signed_request_string)

    likes = []
    if graph:
        likes = graph.get('me/likes')['data']

    context = {'signed_request': signed_request, 'likes': likes}

    return render(request, 'django_facebook/canvas.html', context)


@facebook_required_lazy(page_tab=True)
def page_tab(request, graph):
    '''
    Example of a simple page tab
    '''
    # get signed_request

    signed_request_string = request.POST['signed_request']
    signed_request = parse_signed_request(signed_request_string)
    context = {'signed_request': signed_request}

    return render(request, 'django_facebook/page_tab.html', context)


@facebook_required(scope='publish_stream')
@csrf_protect
def wall_post(request, graph):
    message = request.POST.get('message')
    if message:
        graph.set('me/feed', message=message)
        messages.info(request, 'Posted the message to your wall')
        return next_redirect(request)


@facebook_required(scope=['user_status', 'friends_status'])
def checkins(request, graph):
    '''
    See the facebook docs:
    https://developers.facebook.com/docs/reference/fql/location_post/
    https://developers.facebook.com/docs/reference/api/checkin/

    TODO:
    Add a nice template to this
    '''
    checkins = graph.get('me/checkins')

    raise Exception(checkins)


@facebook_required(scope='publish_stream,user_photos')
@csrf_protect
def image_upload(request):
    fb = get_persistent_graph(request)
    pictures = request.POST.getlist('pictures')

    if pictures:
        for picture in pictures:
            fb.set('me/photos', url=picture, message='the writing is one The '
                   'wall image %s' % picture)

        messages.info(request, 'The images have been added to your profile!')

        return next_redirect(request)


@facebook_required(scope='publish_actions')
@csrf_protect
def open_graph_beta(request):
    '''
    Simple example on how to do open graph postings
    '''
    message = request.POST.get('message')
    if message:
        fb = get_persistent_graph(request)
        entity_url = 'http://www.fashiolista.com/item/2081202/'
        fb.set('me/fashiolista:love', item=entity_url, message=message)
        messages.info(request,
                      'Frictionless sharing to open graph beta action '
                      'fashiolista:love with item_url %s, this url contains '
                      'open graph data which Facebook scrapes' % entity_url)


@facebook_required(scope='publish_actions')
@csrf_protect
def remove_og_share(request):
    from django_facebook.models import OpenGraphShare
    graph = get_persistent_graph(request)
    og_share_id = request.POST.get('og_share_id')
    if og_share_id:
        shares = OpenGraphShare.objects.filter(id=og_share_id)
        for share in shares:
            share.remove(graph)



from django_facebook.management.commands.base import CustomBaseCommand
from django_facebook.utils import queryset_iterator
from optparse import make_option
import datetime


class ExtendTokensCommand(CustomBaseCommand):
    help = 'Extend all the users access tokens\'s, per hour'
    option_list = CustomBaseCommand.option_list + (
        make_option('--all',
                    action='store_true',
                    dest='all',
                    default=False,
                    help='Extend all of them at once'
                    ),
    )

    def handle(self, *args, **kwargs):

        queryset_iterator





user_data = {}

base_data = dict(image='http://u.fashiocdn.com/images/users/0/0/6/O/J/5.jpg',
                 image_thumb='http://u.fashiocdn.com/images/users/0/0/6/O/J/5.jpg')
user_data['no_email'] = {'first_name': 'Thierry', 'last_name': 'Schellenbach', 'verified': True, 'name': 'Thierry Schellenbach', 'gender': 'man', 'image': 'http://graph.facebook.com/me/picture?type=large', 'updated_time':
                         '2010-04-01T14:26:55+0000', 'birthday': '04/07', 'link': 'http://www.facebook.com/profile.php?id=1225707780', 'location': {'id': None, 'name': None}, 'timezone': -5, 'id': '1225707780', 'image_thumb': 'http://graph.facebook.com/me/picture'}
user_data[
    'tschellenbach'] = {'email': 'fake@mellowmorning.com', 'first_name': 'Thierry', 'last_name': 'Schellenbach', 'verified': True, 'name': 'Thierry Schellenbach', 'gender': 'man', 'image': 'http://graph.facebook.com/me/picture?type=large',
                        'updated_time': '2010-04-01T14:26:55+0000', 'birthday': '04/07', 'link': 'http://www.facebook.com/profile.php?id=1225707780', 'location': {'id': None, 'name': None}, 'timezone': -5, 'id': '1225707780', 'image_thumb': 'http://graph.facebook.com/me/picture'}
user_data[
    'new_user'] = {'email': 'fake_new@mellowmorning.com', 'first_name': 'Thierry', 'last_name': 'Schellenbach', 'verified': True, 'name': 'Thierry Schellenbach', 'gender': 'male', 'image': 'http://graph.facebook.com/me/picture?type=large',
                   'updated_time': '2010-04-01T14:26:55+0000', 'birthday': '04/07', 'link': 'http://www.facebook.com/profile.php?id=1225707780', 'location': {'id': None, 'name': None}, 'timezone': -5, 'id': '1225707781', 'image_thumb': 'http://graph.facebook.com/me/picture'}
user_data[
    'no_birthday'] = {'website': 'www.pytell.com', 'first_name': 'Jonathan', 'last_name': 'Pytell', 'verified': True, 'name': 'Jonathan Pytell', 'image': 'http://graph.facebook.com/me/picture?type=large',
                      'updated_time': '2010-01-01T18:13:17+0000', 'link': 'http://www.facebook.com/jpytell', 'location': {'id': None, 'name': None}, 'timezone': -4, 'id': '776872663', 'image_thumb': 'http://graph.facebook.com/me/picture'}
user_data['partial_birthday'] = {'first_name': 'Thierry', 'last_name': 'Schellenbach', 'verified': True, 'name': 'Thierry Schellenbach', 'gender': 'man', 'image': 'http://graph.facebook.com/me/picture?type=large', 'updated_time':
                                 '2010-04-01T14:26:55+0000', 'birthday': '04/07', 'link': 'http://www.facebook.com/profile.php?id=1225707780', 'location': {'id': None, 'name': None}, 'timezone': -5, 'id': '1225707780', 'image_thumb': 'http://graph.facebook.com/me/picture'}
user_data[
    'short_username'] = {'email': 'f@mellowmorning.com', 'first_name': 'Thierry', 'last_name': 'Schellenbach', 'verified': True, 'name': 'Thierry Schellenbach', 'gender': 'man', 'image': 'http://graph.facebook.com/me/picture?type=large',
                         'updated_time': '2010-04-01T14:26:55+0000', 'birthday': '04/07', 'link': 'http://www.facebook.com/profile.php?id=1225707780', 'location': {'id': None, 'name': None}, 'timezone': -5, 'id': '1225707782', 'image_thumb': 'http://graph.facebook.com/me/picture'}
user_data[
    'long_username'] = {'email': 'superduperlongusernamemorethan30characters@mellowmorning.com', 'first_name': 'Thierry123456789123456789123456789123456789123456789123456789123456789123456789123456789', 'last_name': 'Schellenbach123456789123456789123456789123456789123456789123456789', 'verified': True, 'name': 'Thierry Schellenbach',
                        'gender': 'man', 'image': 'http://graph.facebook.com/me/picture?type=large', 'updated_time': '2010-04-01T14:26:55+0000', 'birthday': '04/07', 'link': 'http://www.facebook.com/profile.php?id=1225707780', 'location': {'id': None, 'name': None}, 'timezone': -5, 'id': '1225707782', 'image_thumb': 'http://graph.facebook.com/me/picture'}

user_data['same_username'] = user_data['short_username'].copy()
user_data['same_username'].update(dict(id='1111111', email='t@fake.com'))

user_data['paul'] = {"website": "http://usabilla.com", "last_name": "Veugen", "locale": "en_US", "image": "https://graph.facebook.com/me/picture?access_token=aa&type=large", "quotes": "\"Make no small plans for they have no power to stir the soul.\"  - Machiavelli", "timezone": 1, "education": [{"school": {"id": "114174235260224", "name": "Tilburg University"}, "type": "College", "year": {"id": "136328419721520", "name": "2009"}}, {"school": {"id": "111600322189438", "name": "Tilburg University"}, "type": "Graduate School", "concentration": [{"id": "187998924566551", "name": "Business Communication and Digital Media"}, {"id": "105587479474873", "name": "Entrepreneurship"}], "degree": {"id": "110198449002528", "name": "BA"}}], "id": "555227728", "first_name": "Paul", "verified": True, "sports": [{"id": "114031331940797", "name": "Cycling"}, {"id": "107496599279538", "name": "Snowboarding"}, {"id": "109368782422374", "name": "Running"}], "languages": [{"id": "106059522759137", "name": "English"}, {"id": "107672419256005", "name": "Dutch"}], "location": {
    "id": "111777152182368", "name": "Amsterdam, Netherlands"}, "image_thumb": "https://graph.facebook.com/me/picture?access_token=aa", "email": "paulv@email.com", "username": "pveugen", "bio": "Founder Usabilla. User Experience designer, entrepeneur and digital media addict. On a mission to make user experience research simple and fun with Usabilla.com.", "birthday": "07/11/1984", "link": "http://www.facebook.com/pveugen", "name": "Paul Veugen", "gender": "male", "work": [{"description": "http://usabilla.com", "employer": {"id": "235754672838", "name": "Usabilla"}, "location": {"id": "111777152182368", "name": "Amsterdam, Netherlands"}, "position": {"id": "130899310279007", "name": "Founder, CEO"}, "with": [{"id": "1349358449", "name": "Marc van Agteren"}], "start_date": "2009-01"}, {"position": {"id": "143357809027330", "name": "Communication Designer"}, "start_date": "2000-02", "location": {"id": "112258135452858", "name": "Venray"}, "end_date": "2010-01", "employer": {"id": "133305113393672", "name": "Flyers Internet Communicatie"}}], "updated_time": "2011-11-27T14:34:34+0000"}

user_data['paul2'] = user_data['paul']
user_data[
    'unicode_string'] = {u'bio': u'ATELIER : TODOS LOS DIAS DE 14 A 22 HS C CITA PREVIA. MIX DE ESTETICAS - ECLECTICA - PRENDAS UNICAS-HAND MADE- SEASSON LESS RETROMODERNIDAD -CUSTOMIZADO !!!\r\n\r\nMe encanta todo el proceso que lleva lograr un objeto que sea unico....personalizarlo de acuerdo a cada uno...a cada persona....a su forma de ser.......elegir los materiales, los avios, las telas...siendo la premisa DIFERENCIARSE!!!!\r\n Atemporal..Retro modernidad...Eclecticismo!!!',
                         u'birthday': u'04/09/1973',
                         u'education': [{u'classes': [{u'description': u'direccion de arte',
                                                       u'id': u'188729917815843',
                                                       u'name': u'AAAP'},
                                                      {u'id': u'200142039998266',
                                                       u'name': u'Indumentaria',
                                                       u'with': [{u'id': u'546455670',
                                                                  u'name': u'Gustavo Lento'}]},
                                                      {u'description': u'dise\xf1o ',
                                                       u'id': u'102056769875267',
                                                       u'name': u'figurin',
                                                       u'with': [{u'id': u'1285841477',
                                                                  u'name': u'La Maison Madrid'}]},
                                                      {u'id': u'194463713907701',
                                                       u'name': u'indumentaria -figurin-seriado'},
                                                      {u'id': u'180792848632395',
                                                       u'name': u'indumentaria-figurin -seriado.etc',
                                                       u'with': [{u'id': u'704003068',
                                                                  u'name': u'Mariano Toledo'}]}],
                                         u'concentration': [{u'id': u'176664472378719',
                                                             u'name': u'indumentaria y dise\xf1o textil'}],
                                         u'school': {u'id': u'115805661780893',
                                                     u'name': u'uba arquitectura dise\xf1o y urbanismo'},
                                         u'type': u'College'},
                                        {u'classes': [{u'id': u'116042535136769',
                                                       u'name': u'whadrove asesory'}],
                                         u'concentration': [{u'id': u'146715355391868',
                                                             u'name': u'producer-  stylist -  asesora de imagen -personal-shooper -whadrove asesory'}],
                                         u'school': {u'id': u'107827585907242',
                                                     u'name': u'Miami International University of Art & Design'},
                                         u'type': u'Graduate School'},
                                        {u'classes': [{u'id': u'199608053396849',
                                                       u'name': u'Reporter, Producer and Director'},
                                                      {u'id': u'209828275699578',
                                                       u'name': u'personal shooper'},
                                                      {u'id': u'208913322452336',
                                                       u'name': u'Designer-stylisti'}],
                                            u'concentration': [{u'id': u'184808144897752',
                                                                u'name': u'whadrobing'}],
                                            u'school': {u'id': u'105548239478812',
                                                        u'name': u'New York Institute of Technology'},
                                            u'type': u'College'},
                                        {u'school': {u'id': u'106431826058746',
                                                     u'name': u'Colegio Nuestra Se\xf1ora de la Misericordia'},
                                         u'type': u'High School'},
                                        {u'school': {u'id': u'106011862771707',
                                                     u'name': u'Mary E. Graham'},
                                         u'type': u'High School'},
                                        {u'school': {u'id': u'106011862771707',
                                                     u'name': u'Mary E. Graham'},
                                         u'type': u'High School'}],
                         u'email': u'notfersharerringemail@hotmail.com',
                         u'first_name': u'Fernanda',
                         u'gender': u'female',
                         u'hometown': {u'id': u'109238842431095',
                                       u'name': u'La Coru\xf1a, Galicia, Spain'},
                         u'id': u'1157872766',
                         u'image': u'https://graph.facebook.com/me/picture?type3Dlarge&access_token3D100314226679773|2.1ZkFz1Cusu5RlY1xGft_Pg__.86400.1303534800.4-1157872766|ZgsRrRYwp2pqHEHtncqS5-rBiSg',
                         u'image_thumb': u'https://graph.facebook.com/me/picture?access_token3D100314226679773|2.1ZkFz1Cusu5RlY1xGft_Pg__.86400.1303534800.4-1157872766|ZgsRrRYwp2pqHEHtncqS5-rBiSg',
                         u'languages': [{u'id': u'111021838921932', u'name': u'Espa\xf1ol'},
                                        {u'id': u'110867825605119',
                                         u'name': u'Ingles'},
                                        {u'id': u'108083115891989',
                                         u'name': u'Portugu\xe9s'},
                                        {u'id': u'113051562042989', u'name': u'Italiano'}],
                         u'last_name': u'Ferrer Vazquez',
                         u'link': u'http://www.facebook.com/FernandaFerrerVazquez',
                         u'locale': u'es_LA',
                         u'location': {u'id': u'106423786059675', u'name': u'Buenos Aires, Argentina'},
                         u'name': u'Fernanda Ferrer Vazquez',
                         u'timezone': -3,
                         u'updated_time': u'2011-04-22T04:28:27+0000',
                         u'username': u'FernandaFerrerVazquez',
                         u'verified': True,
                         u'website': u'http://fernandaferrervazquez.blogspot.com/\r\nhttp://twitter.com/fferrervazquez\r\nhttp://comunidad.redfashion.es/profile/fernandaferrervazquez\r\nhttp://www.facebook.com/group.php?gid3D40257259997&ref3Dts\r\nhttp://fernandaferrervazquez.spaces.live.com/blog/cns!EDCBAC31EE9D9A0C!326.trak\r\nhttp://www.linkedin.com/myprofile?trk3Dhb_pro\r\nhttp://www.youtube.com/account#profile\r\nhttp://www.flickr.com/\r\n Mi galer\xeda\r\nhttp://www.flickr.com/photos/wwwfernandaferrervazquez-showroomrecoletacom/ \r\n\r\nhttp://www.facebook.com/pages/Buenos-Aires-Argentina/Fernanda-F-Showroom-Recoleta/200218353804?ref3Dts\r\nhttp://fernandaferrervazquez.wordpress.com/wp-admin/',
                         u'work': [{u'description': u'ATELIER : un mix de esteticas que la hacen eclectica, multicultural, excentrica, customizada, artesanal, unica , exclusiva, lujosa, eco-reciclada,con bordados a mano y texturas..especial atencion a los peque\xf1os detalles!!!! \nSeason less \nHand made ',
                                    u'employer': {u'id': u'113960871952651',
                                                  u'name': u'fernanda ferrer vazquez'},
                                    u'location': {u'id': u'106423786059675',
                                                  u'name': u'Buenos Aires, Argentina'},
                                    u'position': {u'id': u'145270362165639',
                                                  u'name': u'Dise\xf1adora'},
                                    u'projects': [{u'description': u'produccion-estilismo-fotografia-make up-pelo-asesoramiento de imagen-whadrobing-perso\xf1an shooper-rent an outfit!!',
                                                   u'id': u'200218353804',
                                                   u'name': u'Fernanda F -Showroom Recoleta'}],
                                    u'start_date': u'2008-12'},
                                   {u'employer': {u'id': u'198034683561689',
                                                  u'name': u'el atelier de la isla'}}]}

for k, v in user_data.items():
    v.update(base_data)



"""
Forms and validation code for user registration.

"""

from django import forms
from django.utils.translation import ugettext_lazy as _

from django_facebook.utils import get_user_model

attrs_dict = {'class': 'required'}


class FacebookRegistrationFormUniqueEmail(forms.Form):

    """
    Some basic validation, adapted from django registration
    """
    username = forms.RegexField(regex=r'^\w+$',
                                max_length=30,
                                widget=forms.TextInput(attrs=attrs_dict),
                                label=_("Username"),
                                error_messages={'invalid': _("This value must contain only letters, numbers and underscores.")})
    email = forms.EmailField(widget=forms.TextInput(attrs=dict(attrs_dict,
                                                               maxlength=75)),
                             label=_("Email address"))
    password1 = forms.CharField(
        widget=forms.PasswordInput(attrs=attrs_dict, render_value=False),
        label=_("Password"))
    password2 = forms.CharField(
        widget=forms.PasswordInput(attrs=attrs_dict, render_value=False),
        label=_("Password (again)"))

    def clean_username(self):
        """
        Validate that the username is alphanumeric and is not already
        in use.

        """
        try:
            get_user_model().objects.get(
                username__iexact=self.cleaned_data['username'])
        except get_user_model().DoesNotExist:
            return self.cleaned_data['username']
        raise forms.ValidationError(
            _("A user with that username already exists."))

    def clean(self):
        """
        Verifiy that the values entered into the two password fields
        match. Note that an error here will end up in
        ``non_field_errors()`` because it doesn't apply to a single
        field.
        """
        if 'password1' in self.cleaned_data and 'password2' in self.cleaned_data:
            if self.cleaned_data['password1'] != self.cleaned_data['password2']:
                raise forms.ValidationError(
                    _("The two password fields didn't match."))
        return self.cleaned_data

    def clean_email(self):
        """
        Validate that the supplied email address is unique for the
        site.
        """
        if get_user_model().objects.filter(email__iexact=self.cleaned_data['email']):
            raise forms.ValidationError(_(
                "This email address is already in use. Please supply a different email address."))
        return self.cleaned_data['email']



from django.shortcuts import render
from django_facebook.utils import get_registration_backend
from django_facebook.connect import CONNECT_ACTIONS


def register(request):
    """
    A very simplistic register view
    """
    backend = get_registration_backend()
    form_class = backend.get_form_class(request)
    template_name = backend.get_registration_template()

    if request.method == 'POST':
        form = form_class(data=request.POST, files=request.FILES)
        if form.is_valid():
            new_user = backend.register(request,
                                        form=form, **form.cleaned_data)
            # keep the post behaviour exactly the same as django facebook's
            # connect flow
            response = backend.post_connect(
                request, new_user, CONNECT_ACTIONS.REGISTER)
            return response
    else:
        form = form_class()

    context = {'form': form}
    response = render(request, template_name, context)

    return response



# -*- coding: utf-8 -*-

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from django.contrib.auth import logout
from django.utils import six

from open_facebook.api import FacebookAuthorization, OpenFacebook

from django_facebook import settings
from django_facebook.canvas import generate_oauth_url
from django_facebook.connect import connect_user
from django_facebook.exceptions import MissingPermissionsError
from django_facebook.utils import get_profile, ScriptRedirect


class FacebookCanvasMiddleWare(object):

    def process_request(self, request):
        """
        This middleware authenticates the facebook user when
        he/she is accessing the app from facebook (not an internal page)
        The flow is show below:

        if referer is facebook:
            it's a canvas app and the first hit on the app
            If error:
                attempt to reauthenticate using authorization dialog
            if signed_request not sent or does not have the user_id and the access_token:
                user has not authorized app
                redirect to authorization dialog
            else:
                check permissions
                if user is authenticated (in django):
                    check if current facebook user is the same that is authenticated
                    if not: logout authenticated user
                if user is not authenticated:
                    connect_user (django_facebook.connect module)
                changed method to GET. Facebook always sends a POST first.
        else:
            It's an internal page.
            No signed_request is sent.
            Return
        """

        # This call cannot be global'ized or Django will return an empty response
        # after the first one
        redirect_login_oauth = ScriptRedirect(redirect_to=generate_oauth_url(),
                                              show_body=False)
        # check referer to see if this is the first access
        # or it's part of navigation in app
        # facebook always sends a POST reuqest
        referer = request.META.get('HTTP_REFERER', None)
        if referer:
            urlparsed = urlparse(referer)
            is_facebook = urlparsed.netloc.endswith('facebook.com')
            # facebook redirect
            if is_facebook and urlparsed.path.endswith('/l.php'):
                return
            if not is_facebook:
                return
            # when there is an error, we attempt to allow user to
            # reauthenticate
            if 'error' in request.GET:
                return redirect_login_oauth
        else:
            return

        # get signed_request
        signed_request = request.POST.get('signed_request', None)
        try:
            # get signed_request
            parsed_signed_request = FacebookAuthorization.parse_signed_data(
                signed_request)
            access_token = parsed_signed_request['oauth_token']
            if six.PY2:
                facebook_id = long(parsed_signed_request['user_id'])
            else:
                facebook_id = int(parsed_signed_request['user_id'])
        except:
            # redirect to authorization dialog
            # if app not authorized by user
            return redirect_login_oauth
        # check for permissions
        try:
            graph = self.check_permissions(access_token)
        except MissingPermissionsError:
            return redirect_login_oauth
        # check if user authenticated and if it's the same
        if request.user.is_authenticated():
            self.check_django_facebook_user(request, facebook_id, access_token)
        request.facebook = graph
        if not request.user.is_authenticated():
            _action, _user = connect_user(request, access_token, graph)
        # override http method, since this actually is a GET
        if request.method == 'POST':
            request.method = 'GET'
        return

    def check_permissions(self, access_token):
        graph = OpenFacebook(access_token)
        permissions = set(graph.permissions())
        scope_list = set(settings.FACEBOOK_DEFAULT_SCOPE)
        missing_perms = scope_list - permissions
        if missing_perms:
            permissions_string = ', '.join(missing_perms)
            error_format = 'Permissions Missing: %s'
            raise MissingPermissionsError(error_format % permissions_string)

        return graph

    def check_django_facebook_user(self, request, facebook_id, access_token):
        try:
            current_user = get_profile(request.user)
        except:
            current_facebook_id = None
        else:
            current_facebook_id = current_user.facebook_id
        if not current_facebook_id or current_facebook_id != facebook_id:
            logout(request)
            # clear possible caches
            if hasattr(request, 'facebook'):
                del request.facebook
            if request.session.get('graph', None):
                del request.session['graph']
        else:
            # save last access_token to make sure we always have the most
            # recent one
            current_user.access_token = access_token
            current_user.save()



# -*- coding: utf-8 -*-

import datetime
import logging
import unittest

from django.utils.six import StringIO

import mock
from open_facebook.api import *
from open_facebook.exceptions import OpenGraphException
from open_facebook.utils import json
from pprint import pprint

try:
    # python 2 imports
    from urllib2 import HTTPError
except ImportError:
    # python 3 imports
    from urllib.error import HTTPError

logger = logging.getLogger()

TEST_USER_FORCE_CREATE = False
TEST_USER_DICT = {
    'tommy': dict(name='Tommaso Ilgubrab'),
    'thi': dict(name='Thierry Hcabnellehcs'),
    'guy': dict(name='Guyon Eerom', permissions=['read_stream'])
}
TEST_USER_NAMES = [v['name'] for k, v in TEST_USER_DICT.items()]

TEST_USER_OBJECTS = None


def setup_users():
    '''
    Since this is soo slow we only do this once for all tests
    '''
    # caching because these apis are just too damn slow for test driven
    # development
    from django.core.cache import cache
    global TEST_USER_OBJECTS
    if TEST_USER_OBJECTS is None:
        key = 'test_user_objects'
        user_objects = cache.get(key)
        if not user_objects or TEST_USER_FORCE_CREATE:
            logger.info('test user cache not found, rebuilding')
            user_objects = {}
            app_token = FacebookAuthorization.get_app_access_token()
            for user_slug, user_dict in TEST_USER_DICT.items():
                test_user = FacebookAuthorization.get_or_create_test_user(
                    app_token, name=user_dict[
                        'name'], force_create=TEST_USER_FORCE_CREATE,
                    permissions=user_dict.get('permissions')
                )
                user_objects[user_slug] = test_user
            cache.set(key, user_objects, 60 * 60)
        TEST_USER_OBJECTS = user_objects
    return TEST_USER_OBJECTS


class OpenFacebookTest(unittest.TestCase):

    def setUp(self):
        setup_users()
        for user_slug, user_object in TEST_USER_OBJECTS.items():
            setattr(self, user_slug, user_object)

        # capture print statements
        import sys
        self.prints = sys.stdout = StringIO()

    def tearDown(self):
        # complain about print statements
        self.prints.seek(0)
        content = self.prints.read()
        if content:
            raise ValueError('print statement found, output %s' % content)


class TestErrorMapping(OpenFacebookTest):

    def test_syntax_error(self):
        error_response = '''
        {
           'error': {
               'message': 'Syntax error "Expected end of string instead of "?"." at character 14: third_party_id?access_token=AAABbPDnY390BAOZA22ugLfCyr2OGH0k82VJMPJRR8qxceV96nBra53R5ISiou7VOD9eBd21ZCzPZC5Vn1hWbVkY9Qvx9g8wl1NCmuL9vwZDZD',
               'code': 2500,
               'type': 'OAuthException'
           }
        }
        '''
        return

    def test_oauth_errors(self):
        expires_response = '''{
          "error": {
            "type": "OAuthException",
            "message": "Session has expired at unix time SOME_TIME. The current unix time is SOME_TIME."
          }
        } '''
        changed_password_response = '''
        {
          "error": {
            "type": "OAuthException",
            "message": "The session has been invalidated because the user has changed the password."
          }
        }
        '''
        deauthorized_response = '''
        {
          "error": {
            "type": "OAuthException",
            "message": "Error validating access token: USER_ID has not authorized application APP_ID"
          }
        }
        '''
        loggedout_response = '''
        {
          "error": {
            "type": "OAuthException",
            "message": "Error validating access token: The session is invalid because the user logged out."
           }
        }
        '''
        responses = [expires_response, changed_password_response,
                     deauthorized_response, loggedout_response]
        response_objects = []
        for response_string in responses:
            response = json.loads(response_string)

            response_objects.append(response)

        from open_facebook import exceptions as open_facebook_exceptions
        for response in response_objects:
            oauth = False
            try:
                FacebookConnection.raise_error(response['error']['type'],
                                               response['error']['message'])
            except open_facebook_exceptions.OAuthException as e:
                oauth = True
            assert oauth, 'response %s didnt raise oauth error' % response

    def test_non_oauth_errors(self):
        object_open_graph_error = '''
        {"error":
            {"message": "(#3502) Object at URL http://www.fashiolista.com/my_style/list/441276/?og=active&utm_campaign=facebook_action_comment&utm_medium=facebook&utm_source=facebook has og:type of 'website'. The property 'list' requires an object of og:type 'fashiolista:list'. ",
            "code": 3502, "type": "OAuthException"
            }
        }
        '''
        response = json.loads(object_open_graph_error)

        def test():
            FacebookConnection.raise_error(
                response['error']['type'],
                response['error']['message'],
                response['error'].get('code')
            )
        self.assertRaises(OpenGraphException, test)


class Test500Detection(OpenFacebookTest):

    def test_application_error(self):
        '''
        Facebook errors often look like 500s
        Its a silly system, but we need to support it
        This is actually an application error

        '''
        graph = self.guy.graph()

        with mock.patch('urllib2.build_opener') as patched:
            opener = mock.MagicMock()
            response = StringIO('''{
              "error": {
                "type": "OAuthException",
                "message": "Error validating access token: USER_ID has not authorized application APP_ID"
              }
            }''')
            opener.open.side_effect = HTTPError(
                'bla', 500, 'bla', 'bla', response)

            patched.return_value = opener

            def make_request():
                graph.get('me')

            self.assertRaises(facebook_exceptions.OAuthException, make_request)

    def test_facebook_down(self):
        '''
        Facebook errors often look like 500s

        After 3 attempts while facebook is down we raise a FacebookUnreachable
        Exception

        '''
        graph = self.guy.graph()

        with mock.patch('urllib2.build_opener') as patched:
            opener = mock.MagicMock()

            def side_effect(*args, **kwargs):
                response = StringIO(u'''
                <title>Facebook | Error</title>
                Sorry, something went wrong.
                ''')
                http_exception = HTTPError('bla', 505, 'bla', 'bla', response)
                raise http_exception

            opener.open.side_effect = side_effect

            patched.return_value = opener

            def make_request():
                graph.get('me')

            self.assertRaises(
                facebook_exceptions.FacebookUnreachable, make_request)


class TestPublishing(OpenFacebookTest):

    def test_permissions(self):
        graph = self.thi.graph()
        permission_responses = [
            (
                {u'paging': {u'next': u'https://graph.facebook.com/100005270323705/permissions?access_token=CAADD9tTuZCZBQBALXBfM0xDzsn68jAS8HgUSnbhRkZAp5L1FFpY7iLu3aAytCv8jGN4ZCXZAbZCehSvnK7e8d9P22FZCeHarRnFbFne8MluM0S7UNhoCwKWBNrazrs2tjZCIelQAdzesschwzUr3kRCR0oL9bW4Tp6syWmjm0FOUjwZDZD&limit=5000&offset=5000'}, u'data': [
                    {u'user_photos': 1, u'publish_actions': 1, u'read_stream': 1, u'video_upload': 1, u'installed': 1, u'offline_access': 1, u'create_note': 1, u'publish_stream': 1, u'photo_upload': 1, u'share_item': 1, u'status_update': 1}]},
                {u'user_photos': True, u'publish_actions': True, u'read_stream': True, u'video_upload': True, u'installed': True, u'offline_access': True, u'create_note': True, u'publish_stream': True, u'photo_upload': True, u'share_item': True, u'status_update': True}),
            (
                {u'paging': {
                    u'next': u'https://graph.facebook.com/100005270323705/permissions?access_token=CAADD9tTuZCZBQBALXBfM0xDzsn68jAS8HgUSnbhRkZAp5L1FFpY7iLu3aAytCv8jGN4ZCXZAbZCehSvnK7e8d9P22FZCeHarRnFbFne8MluM0S7UNhoCwKWBNrazrs2tjZCIelQAdzesschwzUr3kRCR0oL9bW4Tp6syWmjm0FOUjwZDZD&limit=5000&offset=5000'}, u'data': []},
                {}),
        ]
        # test the full flow, just check no errors are raised
        live_permissions = graph.permissions()
        # test weird responses
        for response, correct_permissions in permission_responses:
            with mock.patch('open_facebook.api.OpenFacebook.get') as g:
                g.return_value = response
                permissions = graph.permissions()
                self.assertEqual(permissions, correct_permissions)

    def test_wallpost(self):
        graph = self.thi.graph()
        now = datetime.datetime.now()
        result = graph.set('me/feed', message='This should work %s' % now)
        self.assertTrue(result['id'])
        graph.delete(result['id'])

        # we have no permissions, this should fail
        guy_graph = self.guy.graph()
        try:
            guy_graph.set('me/feed', message='Nonnonono')
            raise ValueError('We were expecting a permissions exception')
        except facebook_exceptions.PermissionException as e:
            pass

    def test_og_follow(self):
        return
        # perform an og follow
        graph = self.thi.graph()
        path = 'me/og.follows'
        result = graph.set(path, profile=self.guy.id)
        self.assertTrue(result['id'])

        # now try removing it
        remove_path = result['id']
        deleted = graph.delete(remove_path)

    def test_og_adjust(self):
        return
        # perform an og follow
        graph = self.thi.graph()
        path = 'me/og.follows'
        result = graph.set(path, profile=self.guy.id)
        self.assertTrue(result['id'])

        change_result = graph.set(result['id'], message='hello world')
        assert change_result is True

    def test_og_explicit_share(self):
        return
        # perform an og follow
        graph = self.thi.graph()
        path = 'me/og.follows'
        result = graph.set(
            path, profile=self.guy.id, fb__explicitly_shared='true')
        self.assertTrue(result['id'])


class TestOpenFacebook(OpenFacebookTest):

    def test_cookie_parsing(self):
        cookie = 'F7cndfQuSIkcVHWIgg_SHQ4LIDJXeeHhiXUNjesOw5g.eyJhbGdvcml0aG0iOiJITUFDLVNIQTI1NiIsImNvZGUiOiJVMTZuMFNoWVUxSTJ5VEFJMVZ0RmlvZTdhRVRaaEZ4cGV5d1hwYnZvOUprLmV5SnBkaUk2SW1OcmFGVXlWR053ZDA1VlMwSTRlUzFzZDA1WmFtY2lmUS5rZl9RTUhCMnVFTVh5YW83UU5UcnFGMlJzOGxxQUxrM1AxYm8zazBLMm5YUXpOZW5LSVlfczBVV3ZNbE1jTXAzcE04TXNLNVVDQUpjWlQ1N1ZaZXFkS3ZPeXRFbmdoODFxTmczTXVDeTBHNjB6WjFBOWZGZlpHenVDejdKSEVSSCIsImlzc3VlZF9hdCI6MTMxMTYwMDEyNywidXNlcl9pZCI6Nzg0Nzg1NDMwfQ'
        parsed_cookie = FacebookAuthorization.parse_signed_data(cookie)
        assert 'code' in parsed_cookie

    def test_code_conversion(self):
        from open_facebook import exceptions as open_facebook_exceptions
        # before testing update this with a valid code, hope facebook comes
        # with a way to automate this
        code = 'AQDByzD95HCaQLIY3PyQFvCJ67bkYx5f692TylEXARQ0p6_XK0mXGRVBU3G759qOIa_A966Wmm-kxxw1GbXkXQiJj0A3b_XNFewFhT8GSro4i9F8b_7q1RSnKzfq327XYno-Qw4NGxm0ordSl0gJ0YTjhwY8TwSMy2b2whD5ZhHvaYkEaC1J-GcBhkF7o4F2-W8'
        # the redirect uri needs to be connected
        try:
            user_token = FacebookAuthorization.convert_code(
                code, redirect_uri='http://local.mellowmorning.com:8080')
            facebook = OpenFacebook(user_token['access_token'])
            facebook.me()
        except open_facebook_exceptions.ParameterException as e:
            pass

    def test_fql(self):
        facebook = self.thi.graph()
        result = facebook.fql('SELECT name FROM user WHERE uid = me()')
        assert 'name' in result[0]

    def test_open_api(self):
        facebook = self.guy.graph()
        assert 'name' in facebook.me()
        assert facebook.get('fashiolista')

    def test_albums(self):
        graph = self.guy.graph()
        graph.version = 'v2.1'
        albums = graph.get('me/albums', fields='id,type')['data']
        album_type_dict = {a['type']: a for a in albums}
        profile_album = album_type_dict.get('profile')
        cover_album = album_type_dict.get('cover')
        if profile_album:
            pictures = graph.get('%s/photos' % profile_album['id'])['data'][:3]
            for picture in pictures:
                print(picture['source'])
        if cover_album:
            pictures = graph.get('%s/photos' % cover_album['id'])['data'][:3]
            for picture in pictures:
                print(picture['source'])



from django.utils.safestring import mark_safe


try:
    unicode = unicode
except NameError:
    unicode = str


def facebook(request):
    context = {}
    from django_facebook import settings as fb_settings
    from open_facebook.utils import json
    context['FACEBOOK_APP_ID'] = fb_settings.FACEBOOK_APP_ID
    context['FACEBOOK_DEFAULT_SCOPE'] = fb_settings.FACEBOOK_DEFAULT_SCOPE

    default_scope_js = unicode(json.dumps(
        fb_settings.FACEBOOK_DEFAULT_SCOPE))
    default_scope_js = mark_safe(default_scope_js)
    context['FACEBOOK_DEFAULT_SCOPE_JS'] = default_scope_js

    return context



from django.test import TestCase
from django.test.client import Client
from django.utils.six import StringIO

from django_facebook.utils import get_user_model, is_user_attribute
from mock import patch


class FacebookTest(TestCase):

    '''
    Normal Facebook tests run against a fake API
    '''

    def setUp(self):
        from django_facebook.test_utils.mocks import MockFacebookAPI, MockFacebookAuthorization, RequestMock
        import sys

        self.prints = sys.stdout = StringIO()

        from open_facebook import api
        import open_facebook

        self.originalAPI = open_facebook.OpenFacebook
        self.originalAuthorization = open_facebook.FacebookAuthorization

        open_facebook.OpenFacebook = api.OpenFacebook = MockFacebookAPI
        open_facebook.FacebookAuthorization = api.FacebookAuthorization = MockFacebookAuthorization

        rf = RequestMock()
        self.request = rf.get('/')
        self.client = Client()

        # time to setup the test user
        user_model = get_user_model()
        user_dict = dict(
            username='tschellenbach',
            is_staff=False,
            is_active=True,
            email="fake@mellowmorning.com",
        )
        test_ip = '127.0.0.1'
        # hack to make fashiolista tests run ok
        if is_user_attribute('registration_ip'):
            user_dict['registration_ip'] = test_ip
            user_dict['last_login_ip'] = test_ip
        user_model.objects.create(**user_dict)

        from django.conf import settings
        if getattr(settings, 'MODE', None) == 'userena':
            from django.core.management import call_command
            call_command('check_permissions', output=False)

    def tearDown(self):
        from open_facebook import api
        import open_facebook
        open_facebook.OpenFacebook = api.OpenFacebook = self.originalAPI
        open_facebook.FacebookAuthorization = api.FacebookAuthorization = self.originalAuthorization

        self.prints.seek(0)
        content = self.prints.read()
        if content:
            raise ValueError('print statement found, output %s' % content)

    def create_patch(self, name, return_value=None):
        '''
        Easy workaround for having to nest mock.patch context managers
        '''
        patcher = patch(name)
        thing = patcher.start()
        if return_value is not None:
            thing.return_value = return_value
        self.addCleanup(patcher.stop)
        return thing

    def mock_authenticated(self):
        '''
        Fake that we are authenticated
        '''
        from django_facebook.test_utils.mocks import MockFacebookAPI, MockFacebookAuthorization, RequestMock
        graph = MockFacebookAPI('short_username')
        self.create_patch('django_facebook.decorators.has_permissions', True)
        self.create_patch(
            'django_facebook.connect.get_facebook_graph', graph)
        self.create_patch(
            'django_facebook.decorators.get_persistent_graph', graph)
        self.create_patch(
            'django_facebook.decorators.require_persistent_graph', graph)


class LiveFacebookTest(TestCase):

    '''
    Live Facebook Tests run against the actual Facebook API
    '''
    pass



from django.contrib import admin
from django.conf import settings
from django.core.urlresolvers import reverse
from django_facebook import admin_actions
from django_facebook import models
from django_facebook import settings as facebook_settings
from django_facebook.utils import get_profile


class FacebookUserAdmin(admin.ModelAdmin):
    list_display = ('user_id', 'name', 'facebook_id',)
    search_fields = ('name',)


class FacebookLikeAdmin(admin.ModelAdmin):
    list_display = ('user_id', 'name', 'category', 'facebook_id',)
    search_fields = ('name',)
    filter_fields = ('category',)


class FacebookProfileAdmin(admin.ModelAdmin):
    list_display = ('image_', 'user', 'facebook_name', 'facebook_id',)

    raw_id_fields = ('user',)

    search_fields = ('facebook_name', 'facebook_id',)

    def image_(self, instance):
        return """<span style="
        background-image: url({0});
        background-size: cover;
        width: 21px;
        height: 21px;
        display: inline-block;
        outline: 1px solid #DDD;
        position: absolute;
        margin-top: -3px;
    "></span>""".format(
            instance.image.url if (instance and instance.image) else ''
        )
    image_.allow_tags = True


def facebook_profile(open_graph_share):
    '''
    Nicely displayed version of the facebook user
    with user id and image and link to facebook :)
    '''
    user = open_graph_share.user
    profile = get_profile(user)
    facebook_id = profile.facebook_id
    facebook_url = 'http://www.facebook.com/%s/' % facebook_id
    link = '<p><a href="%s"><img src="http://graph.facebook.com/%s/picture/?type=large" width="100px" style="float:left"/>%s</a><br/></p>' % (
        facebook_url, facebook_id, facebook_id)
    return link

facebook_profile.allow_tags = True
facebook_profile.short_description = 'Profile'


class OpenGraphShareAdmin(admin.ModelAdmin):
    raw_id_fields = ['user']
    list_display = ['user', 'action_domain', facebook_profile, 'view_share',
                    'completed_at', 'removed_at', 'error_message']
    actions = [admin_actions.retry_open_graph_share,
               admin_actions.retry_open_graph_share_for_user]

    def view_share(self, instance):
        share_id = instance.share_id
        url_format = 'https://developers.facebook.com/tools/explorer/%s/?method=GET&path=%s%%2F'
        url = url_format % (facebook_settings.FACEBOOK_APP_ID, share_id)
        template = '<a href="%s">%s</a>' % (url, share_id)
        return template
    view_share.allow_tags = True


if getattr(settings, 'AUTH_PROFILE_MODULE', None) == 'django_facebook.FacebookProfile':
    admin.site.register(models.FacebookProfile, FacebookProfileAdmin)

admin.site.register(models.FacebookUser, FacebookUserAdmin)
admin.site.register(models.FacebookLike, FacebookLikeAdmin)
admin.site.register(models.OpenGraphShare, OpenGraphShareAdmin)



#!/usr/bin/env python
from distutils.util import convert_path
from django_facebook import __version__, __maintainer__, __email__
from fnmatch import fnmatchcase
import os
import sys
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages


standard_exclude = ['*.py', '*.pyc', '*~', '.*', '*.bak']
standard_exclude_directories = [
    '.*', 'CVS', '_darcs', './build', './docs',
    './dist', 'EGG-INFO', '*.egg-info', 'facebook_profiles'
]


def find_package_data(where='.', package='', exclude=standard_exclude,
                      exclude_directories=standard_exclude_directories,
                      only_in_packages=True, show_ignored=False):
    """
    Return a dictionary suitable for use in ``package_data``
    in a distutils ``setup.py`` file.

    The dictionary looks like::

        {'package': [files]}

    Where ``files`` is a list of all the files in that package that
    don't match anything in ``exclude``.

    If ``only_in_packages`` is true, then top-level directories that
    are not packages won't be included (but directories under packages
    will).

    Directories matching any pattern in ``exclude_directories`` will
    be ignored; by default directories with leading ``.``, ``CVS``,
    and ``_darcs`` will be ignored.

    If ``show_ignored`` is true, then all the files that aren't
    included in package data are shown on stderr (for debugging
    purposes).

    Note patterns use wildcards, or can be exact paths (including
    leading ``./``), and all searching is case-insensitive.
    """

    out = {}
    stack = [(convert_path(where), '', package, only_in_packages)]
    while stack:
        where, prefix, package, only_in_packages = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if os.path.isdir(fn):
                bad_name = False
                for pattern in exclude_directories:
                    if (fnmatchcase(name, pattern)
                            or fn.lower() == pattern.lower()):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, (
                                'Directory %s ignored by pattern %s'
                                % (fn, pattern))
                        break
                if bad_name:
                    continue
                if (os.path.isfile(os.path.join(fn, '__init__.py'))
                        and not prefix):
                    if not package:
                        new_package = name
                    else:
                        new_package = package + '.' + name
                    stack.append((fn, '', new_package, False))
                else:
                    stack.append(
                        (fn, prefix + name + '/', package, only_in_packages))
            elif package or not only_in_packages:
                # is a file
                bad_name = False
                for pattern in exclude:
                    if (fnmatchcase(name, pattern)
                            or fn.lower() == pattern.lower()):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, (
                                'File %s ignored by pattern %s'
                                % (fn, pattern))
                        break
                if bad_name:
                    continue
                out.setdefault(package, []).append(prefix + name)
    return out


excluded_directories = standard_exclude_directories + ['./requirements',
                                                       './scripts']
package_data = find_package_data(exclude_directories=excluded_directories)
license_text = open('LICENSE.txt').read()
long_description = open('README.rest').read()

if 'alpha' in __version__:
    development_status = 'Development Status :: 3 - Alpha'
else:
    development_status = 'Development Status :: 5 - Production/Stable'

CLASSIFIERS = [
    development_status,
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Framework :: Django',
    'Environment :: Web Environment',
]

DESCRIPTION = """Facebook open graph API client in python. Enables django applications to register users using facebook.
Fixes issues with the official but unsupported Facebook python-sdk. Enables mobile facebook authentication.
Canvas page authentication for facebook applications. FQL access via the server side api.
"""

download_url = 'https://github.com/tschellenbach/Django-facebook/archive/v%s.tar.gz' % __version__

setup(
    name='wish-facebook',
    version=__version__,
    url='http://github.com/tschellenbach/Django-facebook',
    author=__maintainer__,
    author_email=__email__,
    license=license_text,
    packages=find_packages(),
    package_data=package_data,
    description=DESCRIPTION,
    long_description=long_description,
    classifiers=CLASSIFIERS,
    install_requires=['unidecode'],
    tests_require=[
        'django',
        'python-memcached',
        'pillow',
        'mock',
        'pytest',
        'pytest-django'
    ],
    test_suite='runtests.runtests',
    zip_safe=False,  # South can't run migrations on zipped eggs.
)



"""
URL patterns for the views included in ``django.contrib.auth``.

Including these URLs (via the ``include()`` directive) will set up the
following patterns based at whatever URL prefix they are included
under:

* User login at ``login/``.

* User logout at ``logout/``.

* The two-step password change at ``password/change/`` and
  ``password/change/done/``.

* The four-step password reset at ``password/reset/``,
  ``password/reset/confirm/``, ``password/reset/complete/`` and
  ``password/reset/done/``.

The default registration backend already has an ``include()`` for
these URLs, so under the default setup it is not necessary to manually
include these views. Other backends may or may not include them;
consult a specific backend's documentation for details.

"""

from django.conf.urls import url
from django.contrib.auth import views as auth_views
from django_facebook import registration_views
from django_facebook.utils import replication_safe

urlpatterns = [
    url(
        r'^login/$',
        replication_safe(auth_views.login),
        {'template_name': 'registration/login.html'},
        name='auth_login'
    ),
    url(
        r'^logout/$',
        replication_safe(auth_views.logout),
        {'template_name': 'registration/logout.html'},
        name='auth_logout'
    ),
    url(
        r'^password/change/$',
        auth_views.password_change,
        name='auth_password_change'
    ),
    url(
        r'^password/change/done/$',
        auth_views.password_change_done,
        name='auth_password_change_done'
    ),
    url(
        r'^password/reset/$',
        auth_views.password_reset,
        name='auth_password_reset'
    ),
    url(
        r'^password/reset/confirm/(?P<uidb36>[0-9A-Za-z]+)-(?P<token>.+)/$',
        auth_views.password_reset_confirm,
        name='auth_password_reset_confirm'
    ),
    url(
        r'^password/reset/complete/$',
        auth_views.password_reset_complete,
        name='auth_password_reset_complete'
    ),
    url(
        r'^password/reset/done/$',
        auth_views.password_reset_done,
        name='auth_password_reset_done'
    ),
    url(
        r'^register/$',
        registration_views.register,
        name='registration_register'
    ),
]



from functools import wraps
import logging
from celery import task
from django.db import IntegrityError
from django_facebook.utils import get_class_for
from django_facebook.signals import facebook_token_extend_finished

logger = logging.getLogger(__name__)


@task.task(ignore_result=True)
def extend_access_token(profile, access_token):
    '''
    Extends the access token to 60 days and saves it on the profile

    :param profile: the profile or user object
    :param access_token: a valid access token
    :type access_token: string
    '''
    results = profile._extend_access_token(access_token)
    return results


@task.task(ignore_result=True)
def store_likes(user, likes):
    '''
    Inserting again will not cause any errors, so this is safe
    for multiple executions

    :param user: The user for which we are storing
    :type user: User object

    :param friends: List of your likes
    :type friends: list
    '''
    converter_class = get_class_for('user_conversion')
    logger.info('celery is storing %s likes' % len(likes))
    converter_class._store_likes(user, likes)
    return likes


@task.task(ignore_result=True)
def store_friends(user, friends):
    '''
    Inserting again will not cause any errors, so this is safe
    for multiple executions

    :param user: The user for which we are storing
    :type user: User object

    :param friends: List of your friends
    :type friends: list
    '''
    converter_class = get_class_for('user_conversion')
    logger.info('celery is storing %s friends' % len(friends))
    converter_class._store_friends(user, friends)
    return friends


@task.task(ignore_result=True)
def get_and_store_likes(user, facebook):
    '''
    Since facebook is quite slow this version also runs the get
    on the background

    Inserting again will not cause any errors, so this is safe
    for multiple executions

    :param user: The user for which we are storing
    :type user: User object

    :param facebook: The graph connection to facebook
    :type facebook: FacebookUserConverter object
    '''
    try:
        logger.info('attempting to get and store friends for %s', user.id)
        stored_likes = facebook._get_and_store_likes(user)
        logger.info('celery is storing %s likes', len(stored_likes))
        return stored_likes
    except IntegrityError as e:
        logger.warn(
            'get_and_store_likes failed for %s with error %s', user.id, e)


@task.task(ignore_result=True)
def get_and_store_friends(user, facebook):
    '''
    Since facebook is quite slow this version also runs the get
    on the background

    Inserting again will not cause any errors, so this is safe
    for multiple executions

    :param user: The user for which we are storing
    :type user: User object

    :param facebook: The graph connection to facebook
    :type facebook: FacebookUserConverter object
    '''
    try:
        logger.info('attempting to get and store friends for %s', user.id)
        stored_friends = facebook._get_and_store_friends(user)
        logger.info('celery is storing %s friends', len(stored_friends))
        return stored_friends
    except IntegrityError as e:
        logger.warn(
            'get_and_store_friends failed for %s with error %s', user.id, e)


@task.task(ignore_result=True)
def remove_share(share):
    '''
    Removes the given open graph share

    :param share: the open graph share object
    '''
    share._remove()


@task.task(ignore_result=True)
def retry_open_graph_share(share, reset_retries=False):
    '''
    We will retry open graph shares after 15m to make sure we dont miss out on any
    shares if Facebook is having a minor outage
    '''
    logger.info('retrying open graph share %s', share)
    share.retry(reset_retries=reset_retries)


@task.task(ignore_result=True)
def retry_open_graph_shares_for_user(user):
    '''
    We retry the open graph shares for a user when he gets a new access token
    '''
    from django_facebook.models import OpenGraphShare
    shares = OpenGraphShare.objects.recently_failed().filter(user=user)[:1000]
    shares = list(shares)
    logger.info('retrying %s shares for user %s', len(shares), user)

    for share in shares:
        retry_open_graph_share(share, reset_retries=True)


def token_extended_connect(sender, user, profile, token_changed, old_token, **kwargs):
    from django_facebook import settings as facebook_settings
    if facebook_settings.FACEBOOK_CELERY_TOKEN_EXTEND:
        # This is only save to run if we are using Celery
        # make sure we don't have troubles caused by replication lag
        retry_open_graph_shares_for_user.apply_async(args=[user], countdown=60)

facebook_token_extend_finished.connect(token_extended_connect)
